// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu
//    mhadji@gmail.com


// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

// example of a Visitor pattern.
// findes the index and leaf IO for answering the query and prints
// the resulting data IDs to stdout.
class MyVisitor : public IVisitor
{
public:
	unsigned long m_indexIO;
	unsigned long m_leafIO;
	vector<byte* > vos;	
	vector<long> aggs;
	vector<long > digestSize;
	vector<long> digPosition;
	vector<long> mcnAggs;
	vector<byte*> mcnDigest;
	vector<long> mcnVOaggs;
	vector<byte*> mcnVOdigest;
	vector<long> mcnAffectedNode;
	long m_agg;

public:
	MyVisitor() : m_indexIO(0), m_leafIO(0) 
	{	
	}

	void visitNode(const INode& n)
	{
		if (n.isLeaf()) m_leafIO++;
		else m_indexIO++;
	}

	void visitData(const IData& d)
	{
		IShape* pS;
		d.getShape(&pS);
			// do something.
		Point key;
		pS->getCenter(key);	
		delete pS;

		long agg=0;
		byte* pData = 0;
		unsigned long cLen = 0;
		d.getData(cLen, &pData);
		memcpy(&agg, pData, cLen);
			// do something.
			//string s = reinterpret_cast<char*>(pData);
			//cout << s << endl;
		delete[] pData;
		
		cout << "Query results: pid key agg "<<endl;
		cout << d.getIdentifier() << " "<< key.m_pCoords[0]<<" "<<agg<<endl;
		m_agg=agg;	
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v)
	{
		cout << v[0]->getIdentifier() << " " << v[1]->getIdentifier() << endl;
	}
	
	void updateVOs(byte** digest, long* agg, long num, long pos) 
	{
		digestSize.insert(digestSize.begin(),num);
		digPosition.insert(digPosition.begin(),pos);
		for (long i=num-1;i>=0;i--)
		{
			byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
			memcpy(dig, digest[i], DIGEST_LENGTH);
			vos.insert(vos.begin(),dig);	
			aggs.insert(aggs.begin(),agg[i]);
		}		
	}
	
	void updateMCN(long agg, byte* digest)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnDigest.insert(mcnDigest.begin(),dig);	
		mcnAggs.insert(mcnAggs.begin(),agg);
	}

	void updateMCNVOs(long agg, byte* digest)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnVOdigest.insert(mcnVOdigest.begin(),dig);	
		mcnVOaggs.insert(mcnVOaggs.begin(),agg);
	}
	
	void updateMCN(long agg, byte* digest, double* pLow, double* pHigh, int dimension)
	{}

	void updateMCNVOs(long agg, byte* digest, double* pLow, double* pHigh, int dimension)
	{}
	
	void updateMCNaffectedNode(long nodeID)
	{
		bool found=false;
		for (int i=0; i<mcnAffectedNode.size(); i++)
		{
			if (mcnAffectedNode[i]==nodeID)
			{
				found=true;
				break;
			}
		}	
		if (!found)
			mcnAffectedNode.insert(mcnAffectedNode.begin(), nodeID);	
		m_indexIO++;
	}

	void computeAggAns()
	{
		long agg=0;
		cout<<"mcnAggs size "<<mcnAggs.size()<<endl;
		for (int i=0;i<mcnAggs.size();i++)
		{
			cout<<mcnAggs[i]<<endl;
			agg+=mcnAggs[i];
		}
		cout<<"Aggregation answer: "<<agg<<endl;
	}
	
	long calculateAggVOsize()
	{
		cout<<"mcnAggs size "<<mcnAggs.size()<<endl;
		cout<<"mcnVOaggs size "<<mcnVOaggs.size()<<endl;
		cout<<"mcnDigest size "<<mcnDigest.size()<<endl;
		cout<<"mcnVOdigest size "<<mcnVOdigest.size()<<endl;
		return (4*(mcnAggs.size()+mcnVOaggs.size())+DIGEST_LENGTH*(mcnDigest.size()+mcnVOdigest.size()));
	}

	void updateMCNcounter(long level)
	{}
};

void verify(vector<long> digestSize, vector<byte*> vos, vector<long> aggs, vector<long> digPosition, long key, long agg, int** label) 
{
	long size=digestSize.size();
	byte* kd=(byte*)malloc(DIGEST_LENGTH);
	byte* input=(byte*)malloc(2*sizeof(long));
	memcpy(input, &key, sizeof(long));
	memcpy(input+sizeof(long), &agg, sizeof(agg));
	SpatialIndex::sha1(input, kd, 2*sizeof(long));		
	byte* lastDigest=kd;
	byte* digests;
	long voSize=vos.size();
	long aggSize=aggs.size();
	long lastAgg=agg;
	long tmpAgg=0;
	*label=(int*)malloc(size*sizeof(int));
	int k=(int)(size-1);
	
	for (long i=0; i< size; i++)
	{
		long s=digestSize.front();
		digestSize.erase(digestSize.begin());
		cout<<"s "<<s<<endl;
		long pos=digPosition.front();
		digPosition.erase(digPosition.begin());
		cout<<"pos "<<pos<<endl;
		(*label)[k--]=(int)pos;
		EVP_MD_CTX tc;
		EVP_MD_CTX_init(&tc);
		EVP_DigestInit_ex(&tc,EVP_sha1(), NULL);
		digests=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		bool meetPos=false;
		long j;
		for (j=0; j<s; j++)
		{
			if (j==pos && meetPos==false)
			{	
				char* d= SpatialIndex::printDigest(lastDigest, DIGEST_LENGTH);			
				for (unsigned long i=0; i< DIGEST_LENGTH*2; i++)
					cout << d[i];
				cout<<"     (computed digest) ";
				delete[] d;
				EVP_DigestUpdate(&tc,lastDigest,DIGEST_LENGTH);
				EVP_DigestUpdate(&tc, &lastAgg, sizeof(long));
				cout<<" "<<lastAgg<<" computed agg"<<endl;
				tmpAgg+=lastAgg;
				delete[] lastDigest;
				meetPos=true;
				j--;
			}
			else
			{
				byte* dig=vos.front();
				vos.erase(vos.begin());
				char* d= SpatialIndex::printDigest(dig, DIGEST_LENGTH);
				for (unsigned long i=0; i< DIGEST_LENGTH*2; i++)
					cout << d[i];
				delete[] d;

				EVP_DigestUpdate(&tc,dig,DIGEST_LENGTH);
				
				long a=aggs.front();
				aggs.erase(aggs.begin());
				byte* bytea=(byte*)malloc(sizeof(long));
				memcpy(bytea, &a, sizeof(long));
				EVP_DigestUpdate(&tc,bytea, sizeof(long));
				tmpAgg+=a;
				cout<<" "<<a<<endl;
				delete[] bytea;
				delete[] dig;
			}
		}	
		if (j==pos && meetPos==false)
		{
			char* d= SpatialIndex::printDigest(lastDigest, DIGEST_LENGTH);			
			for (unsigned long i=0; i< DIGEST_LENGTH*2; i++)
				cout << d[i];
			cout<<"     (computed digest) ";
			delete[] d;
			EVP_DigestUpdate(&tc,lastDigest,DIGEST_LENGTH);
			EVP_DigestUpdate(&tc, &lastAgg, sizeof(long));
			cout<<" "<<lastAgg<<" computed agg"<<endl;
			tmpAgg+=lastAgg;
			delete[] lastDigest;
			meetPos=true;

		}
		EVP_DigestFinal_ex(&tc,digests,NULL);
		EVP_MD_CTX_cleanup(&tc);
		lastDigest=digests;
		lastAgg=tmpAgg;
		tmpAgg=0;
	}
	cout<<"query verify digest:"<<endl;
	char* d= SpatialIndex::printDigest(lastDigest, DIGEST_LENGTH);		
	for (unsigned long i=0; i< DIGEST_LENGTH*2; i++)
		cout << d[i];
	cout<<endl;
	cout<<"VOs size and total length:"<<voSize<<" "<<voSize*DIGEST_LENGTH<<endl;
	cout<<"Aggs size ang total length:"<<aggSize<<" "<<aggSize*sizeof(int)<<endl;
	cout<<"Labels: ";
	for (int i=0; i<(int)size;i++)
		cout<<(*label)[i];
	cout<<endl;
	delete[] d;
	delete[] lastDigest;
}



int main(int argc, char** argv)
{
	try
	{
		if (argc != 5)
		{
			cerr << "Usage: " << argv[0] << " query_file tree_file query_type [pointsearch | rangequery | aggquery] dimension." << endl;
			return -1;
		}

		int queryType = 0;

		if (strcmp(argv[3], "pointsearch") == 0) queryType = 0;
		else if (strcmp(argv[3], "rangequery") == 0) queryType = 1;
		else if (strcmp(argv[3], "aggquery") == 0) queryType = 2;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open query file " << argv[1] << "." << endl;
			return -1;
		}

		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::loadDiskStorageManager(baseName);
			// this will try to locate and open an already existing storage manager.

		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 10, false);
			// applies a main memory random buffer on top of the persistent storage manager
			// (LRU buffer, etc can be created the same way).

		// If we need to open an existing tree stored in the storage manager, we only
		// have to specify the index identifier as follows

		ISpatialIndex* tree = BTree::loadBTree(*file, 1);
		long mcnSize=0;
		long verifyNode=0;
		long mcnVOSize=0;
		
		unsigned long count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long id, op;
		int dimension=atoi(argv[4]);
		double*  l;
		double*  h;
		l=(double*)malloc(dimension*sizeof(double));
		h=(double*)malloc(dimension*sizeof(double));
	
		Tools::ResourceUsage ru;

		//tree->showTree();
		ru.start();
		long communication=0;

		while (fin)
		{
			fin >> op;
			for (int i=0;i<dimension;i++)
				fin >> l[i];
			for (int i=0;i<dimension;i++)
				fin >> h[i];

			if (! fin.good()) continue; // skip newlines, etc.

			if (op == QUERY)
			{

				MyVisitor vis;

				if (queryType == 0)
				{
					Point p = Point(l, 1);
					tree->pointLocationQuery(p, vis);
					int* label;
					verify(vis.digestSize, vis.vos, vis.aggs, vis.digPosition, (long)l[0], vis.m_agg, &label);
				}
				else if (queryType == 1)
				{
					LineSegment r = LineSegment(l, h, 1);
					tree->containsWhatQuery(r, vis);
				}
				else if (queryType == 2)
				{
					LineSegment r = LineSegment(l, h, 1);
					//cout<<"bfe query "<<endl;
					tree->mcnQuery(r, vis);	
					//cout<<"end query "<<endl;
					//vis.computeAggAns();
					long t=vis.calculateAggVOsize();
					cout<<"count="<<count<<" VO + Res size "<<t<<endl;
					communication+=t;
					mcnSize+= vis.mcnAggs.size(); 
					mcnVOSize+= vis.mcnVOaggs.size();
					verifyNode+= vis.mcnAffectedNode.size(); 
					
				}
				indexIO += vis.m_indexIO;
				leafIO += vis.m_leafIO;
					// example of the Visitor pattern usage, for calculating how many nodes
					// were visited.
					}
			else
			{
				cerr << "This is not a query operation." << endl;
			}

			if ((count % 1000) == 0 && count!=0) cerr << count << endl;
			count++;
		}

		ru.stop();

		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Buffer hits: " << file->getHits() << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;
		
		cerr << "communication cost "<<communication/count<<endl;
		cerr << "mcn size " <<mcnSize/count<<endl;	
		cerr << "mcn vo size " <<mcnVOSize/count<<endl;
		cerr << "verify node "<<verifyNode/count<<endl;
		cerr << "ave IO "<< indexIO/count<<endl;

		delete tree;
		delete file;
		delete diskfile;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail writting the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
