// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double* m_dimensionattrs;
	double m_measureattrs;
	double m_onedimensionIndex;
	int m_dimension;
	unsigned long id;

	Point(int dimension)
	{
		m_dimension=dimension;
		m_dimensionattrs=(double*)malloc(dimension*sizeof(double));
	}
	
	void computeOneDimensionIndex(unsigned long domainsize)
	{
		m_onedimensionIndex=0.0;
		for (int i=0; i<m_dimension; i++)
			m_onedimensionIndex+=m_dimensionattrs[i]*pow((double)domainsize, (double)(m_dimension-(i+1)));
	}
	
	~Point()
	{
		//delete[] m_dimensionattrs;
	}
};

int main(int argc, char** argv)
{
	if (argc != 5)
	{
		cerr << "Usage: " << argv[0] << " data_file dimension number_of_query selectivity." << endl;
		return -1;
	}

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open data file " << argv[1] << "." << endl;
		return -1;
	}
	
	int dimension = atoi(argv[2]);
	unsigned long numberOfQuery = atol(argv[3]);
	double selectivity= atof(argv[4]);
	map<unsigned long, Point> data;
	vector<unsigned long> keys;
	Tools::Random rnd;
	int op;

	while (fin)
	{
		Point r=Point(dimension);
		fin >> op >> r.id >> r.m_measureattrs;
		for (int i=0;i<dimension;i++)
			fin >> r.m_dimensionattrs[i];
		fin >> r.m_onedimensionIndex;
		if (! fin.good()) continue; // skip newlines, etc.
		data.insert(pair<unsigned long, Point>((unsigned long)r.m_onedimensionIndex, r));	
		keys.push_back((unsigned long)r.m_onedimensionIndex);
	}
	
	std::sort(keys.begin(),keys.end());

	unsigned long numberOfObjects=keys.size();
	unsigned long range=(unsigned long)ceil(selectivity*numberOfObjects);
	for (unsigned long i = 0; i < numberOfQuery; i++)
	{
		unsigned long start = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects-range-1));
		unsigned long x1 = keys[start];
		unsigned long x2 = keys[start+range-2];
		map<unsigned long, Point>::iterator it1 = data.find(x1);
		assert(it1 != data.end());
		map<unsigned long, Point>::iterator it2 = data.find(x2);
		assert(it2 != data.end());
		cout << QUERY;
		for (int j=0;j<dimension;j++)
			cout <<  " " << (*it1).second.m_dimensionattrs[j] << " " << (*it2).second.m_dimensionattrs[j];
		cout<<endl;
	}

	return 0;
}
