// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double* m_dimensionattrs;
	double m_measureattrs;
	double m_onedimensionIndex;
	int m_dimension;
	unsigned long id;

	Point(int dimension)
	{
		m_dimension=dimension;
		m_dimensionattrs=(double*)malloc(dimension*sizeof(double));
	}
	
	void computeOneDimensionIndex(unsigned long domainsize)
	{
		m_onedimensionIndex=0.0;
		for (int i=0; i<m_dimension; i++)
			m_onedimensionIndex+=m_dimensionattrs[i]*pow((double)domainsize, (double)(m_dimension-(i+1)));
	}
	
	~Point()
	{
		//delete[] m_dimensionattrs;
	}
};

int main(int argc, char** argv)
{
	if (argc != 5)
	{
		cerr << "Usage: " << argv[0] << " dimension domainsize number_of_query selectivity." << endl;
		return -1;
	}
	
	int dimension = atoi(argv[1]);
	unsigned long domainsize = atol(argv[2]);
	unsigned long numberOfQuery = atol(argv[3]);
	double selectivity= atof(argv[4]);
	Tools::Random rnd;
	int op;
	unsigned long range=(unsigned long)floor(selectivity*(double)domainsize);
	unsigned long l,h;

	for (unsigned long i = 0; i < numberOfQuery; i++)
	{
		cout << QUERY;
		for (int j=0;j<dimension;j++)
		{
			l=rnd.nextUniformLong(1, domainsize-range);
			h=l+range;
			cout<<" "<< l <<" "<< h;
		}
		cout<<endl;
	}

	return 0;
}
