// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	long* m_dimensionattrs;
	long m_measureattrs;
	long m_onedimensionIndex;
	int m_dimension;
	long id;

	Point(int dimension)
	{
		m_dimension=dimension;
		m_dimensionattrs=(long*)malloc(dimension*sizeof(long));
	}
	
	void computeOneDimensionIndex(long domainsize)
	{
		m_onedimensionIndex=0;
		for (int i=0; i<m_dimension; i++)
			m_onedimensionIndex+=m_dimensionattrs[i]*(long)pow((double)domainsize, (double)(m_dimension-(i+1)));
	}
	
	~Point()
	{
		//delete[] m_dimensionattrs;
	}
};

bool indexMatched(long* s, long* d, int dimension)
{
	bool match=true;
	for (int i=0; i<dimension;i++)
	{
		if (s[i]!=d[i])
		{
			match=false;
			break;
		}	
	}
	return match;
}

void updateDimensionIndex(long*& d, int dimension, long domainsize)
{
	for (int i=dimension-1;i>=0;i--)
	{
		d[i]=(d[i]+1)%domainsize;
		if (d[i]!=0)
			break;
	} 	
}

int main(int argc, char** argv)
{
	if (argc != 4)
	{
		cerr << "Usage: " << argv[0] << " data_file dimension domainsize." << endl;
		return -1;
	}

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open data file " << argv[1] << "." << endl;
		return -1;
	}
	
	int dimension = atoi(argv[2]);
	long domainsize=atol(argv[3]);
	long N=(long)pow((double)domainsize, (double)dimension);
	long* c=(long*)malloc(N*sizeof(long));
	map<long, Point> data;
	map<long, Point> prefixsum;
	vector<long> keys;
	Tools::Random rnd;
	int op;
	long* dimensionIndex=(long*)malloc(dimension*sizeof(long));
	memset(dimensionIndex, 0, dimension);


	while (fin)
	{
		Point r=Point(dimension);
		fin >> op >> r.id >> r.m_measureattrs;
		for (int i=0;i<dimension;i++)
			fin >> r.m_dimensionattrs[i];
		fin >> r.m_onedimensionIndex;
		if (! fin.good()) continue; // skip newlines, etc.
		
		data.insert(pair<long, Point>((long)r.m_onedimensionIndex, r));	
		keys.push_back((long)r.m_onedimensionIndex);
	}
	map<long, Point>::iterator it;
	vector<long>::iterator itv;
	while(keys.size()>0)
	{
		long pos= static_cast<long>(rnd.nextUniformLong(0,keys.size()));
		it = data.find(keys[pos]);	
		cout << INSERT << " " << (*it).second.id<<" "<<(*it).second.m_measureattrs;
		for (int i=0; i<dimension; i++)
		{
			cout << " " << (*it).second.m_dimensionattrs[i];
		}
		cout << " " << (*it).second.m_onedimensionIndex << endl;
		itv=keys.begin();
		keys.erase(itv+pos);
		data.erase(it);
	}
	return 0;
}
