// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//      lifeifei@cs.bu.edu
//    mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

long computeCapacity(long p)
{
	return (long)ceil((double)p/24.0);
}

int main(int argc, char** argv)
{
	try
	{
		if (argc != 6)
		{
			cerr << "Usage: " << argv[0] << " input_file tree_file pagesize query_type [pointsearch | rangequery] dimension." << endl;
			return -1;
		}

		int queryType = 0;

		if (strcmp(argv[4], "pointsearch") == 0) queryType = 0;
		else if (strcmp(argv[4], "rangequery") == 0) queryType = 1;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		char* treefile = argv[2];
		long pagesize = atol(argv[3]);
		long capacity= computeCapacity(pagesize);
		NwayTree* tree= new NwayTree((int)capacity);
		int dimension = atoi(argv[5]);

		unsigned long count = 0;
		unsigned long id, op;
		long x;
		long* y=(long*)malloc(dimension*sizeof(long));
		long oneDindex;
		Tools::ResourceUsage ru;
		std::vector<long> aggs;
		ru.start();

		while (fin)
		{
			fin >> op >> x ;
			for (int i=0;i<dimension;i++)
				fin >> y[i];
			fin >> oneDindex;
			if (! fin.good()) continue; // skip newlines, etc.

			if (op == INSERT)
			{
				aggs.push_back(x);
				
			}
			else if (op == DELETE)
			{

			}
			else if (op == QUERY)
			{

			}

			//if (((count+1) % 10000) == 0 && count!=0)
				//cerr << count+1 << endl;
			count++;
		}
		//cout << "aggs size "<<aggs.size()<<endl;
		
		long size=aggs.size();
		long* aggset= new long[size];
		for (long i=0;i<size;i++)
		{
			aggset[i]=(long)aggs.front();
			aggs.erase(aggs.begin());
		}
		tree->bulkloadTree(aggset, size,0);	
		//tree->showTree();
		tree->printStatistics();
		long treesize=tree->calculateTreeSize();
		cout<<"fanout= "<<capacity<<endl;
		cout<<"tree size= "<<treesize<<" bytes"<<endl;	
		tree->storeToFile(treefile);
		
		/*delete tree;
			
		tree=new NwayTree(atoi(argv[3]));
		cout<<" i am here"<<endl;
		tree->loadFromFile(treefile);
		cout<<" i am here"<<endl;
		tree->showTree();*/
		/*		
		byte** vos;
		long voSize;
		long* pos;				
		long posSize;
		long searchIndex=55;
		tree->constructVOs(searchIndex, vos, voSize, &pos, posSize);
		cout<< "vo size: " << voSize <<endl;
		for (long i=0; i<voSize; i++)
		{
			char* d= SpatialIndex::printDigest(vos[i], DIGEST_LENGTH);		
			for (long j=0; j< DIGEST_LENGTH*2; j++)
				cout << d[j];
			cout<<endl;
		}
		delete[] pos;
		delete[] vos;


		searchIndex=60;
		tree->constructVOs(searchIndex, vos, voSize, &pos, posSize);
		cout<< "vo size: " << voSize <<endl;
		for (long i=0; i<voSize; i++)
		{
			char* d= SpatialIndex::printDigest(vos[i], DIGEST_LENGTH);		
			for (long j=0; j< DIGEST_LENGTH*2; j++)
				cout << d[j];
			cout<<endl;
		}*/

		ru.stop();
		
		/*long* indexarray=(long*)malloc(2*sizeof(long));
		indexarray[0]=55;
		indexarray[1]=60;
		long numberofhash=0;
		long numberofnodes=0;
		tree->calculateVOsize(indexarray, 2, &numberofhash, &numberofnodes);
		cout<<"number of hashes: "<<numberofhash<<" number of nodes: "<<numberofnodes<<endl;*/

		cerr << "Operations: " << count << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;
	

		delete tree;
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
