// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

long computeCapacity(long p)
{
	return (long)ceil((double)p/24.0);
}

long computeOneDindex(long* x, long domainsize, long dimension)
{
	long oneDindex=0;
	for (long i=0;i<dimension;i++)
		oneDindex+=x[i]*(long)pow((double)domainsize, (double)(dimension-(i+1)));	
	return oneDindex;
}

void buildPSelements(long* x, long* y, long* t, long i, long dimension, long**& psarray, long& curRow)
{
	if (i>=dimension)
	{
		for (long j=0;j<dimension;j++)
			psarray[curRow][j]=t[j];	
		curRow++;
		return;
	}
	for (long j=0;j<2;j++)
	{
		if (j==0)
		{
			t[i]=x[i];
			//cout<<"t "<<i<<" = "<<"x "<<i<<endl;
		}
		else
		{
			t[i]=y[i];
			//cout<<"t "<<i<<" = "<<"y "<<i<<endl;

		}
		buildPSelements(x, y, t, i+1, dimension, psarray, curRow);
	}
}

long calVOsize(long numberofhash, long numberofnodes, long numberofelements)
{
	return (numberofhash*DIGEST_LENGTH+numberofnodes*(4+4)+numberofelements*4);
}
int main(int argc, char** argv)
{
	try
	{
		if (argc != 7)
		{
			cerr << "Usage: " << argv[0] << " input_file tree_file pagesize query_type [pointsearch | rangequery] dimension domainsize." << endl;
			return -1;
		}

		int queryType = 0;

		if (strcmp(argv[4], "pointsearch") == 0) queryType = 0;
		else if (strcmp(argv[4], "rangequery") == 0) queryType = 1;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open data file " << argv[1] << "." << endl;
			return -1;
		}

		char* treefile = argv[2];
		long pagesize = atol(argv[3]);
		long capacity= computeCapacity(pagesize);
		NwayTree* tree= new NwayTree((int)capacity);
		int dimension = atoi(argv[5]);
		long domainsize = atol(argv[6]);
		tree->loadFromFile(treefile);
		//tree->showTree();
		cout<<"after loading tree"<<endl;

		unsigned long count = 0;
		unsigned long id, op;
		long* x=(long*)malloc(dimension*sizeof(long));
		long* y=(long*)malloc(dimension*sizeof(long));
		long oneDindex;
		Tools::ResourceUsage ru;
		std::vector<long> aggs;
		ru.start();
		long numelements=(long)pow(2.0, dimension);
		long* indexarray=(long*)malloc(numelements*sizeof(long));
		long* tmp=(long*)malloc(dimension*sizeof(long));
		long** psarray=(long**)malloc(numelements*sizeof(long*));
		for (long i=0;i<numelements;i++)
			psarray[i]=(long*)malloc(dimension*sizeof(long));
		long totalHash=0;
		long totalNodeVerify=0;

		while (fin)
		{
			fin >> op;
			for (int i=0;i<dimension;i++)
				fin >> x[i];
			for (int i=0;i<dimension;i++)
				fin >> y[i];
			if (! fin.good()) continue; // skip newlines, etc.
			long k=0;
			long curRow=0;
			buildPSelements(x, y, tmp, k, dimension, psarray, curRow);
			cout<<"after buidling ps "<<endl;
			for (long i=0;i<numelements;i++)
			{
				for (long j=0;j<dimension;j++)
					cout<<" "<<psarray[i][j];
				cout<<endl;
				indexarray[i]=computeOneDindex(psarray[i], domainsize, dimension);
			}
			for (long i=0;i<numelements;i++)
				cout<<"indexarray "<<indexarray[i]<<endl;
			if (op == INSERT)
			{
				
			}
			else if (op == DELETE)
			{

			}
			else if (op == QUERY)
			{
				if (queryType==0)
				{
					byte** vos;
					long voSize;
					long* pos;				
					long posSize;
					long searchIndex=indexarray[0];
					cout<<"searching "<<searchIndex<<endl;
					tree->constructVOs(searchIndex, vos, voSize, &pos, posSize);
					cout<< "vo size: " << voSize <<endl;
					for (long i=0; i<voSize; i++)
					{
						char* d= SpatialIndex::printDigest(vos[i], DIGEST_LENGTH);		
						for (long j=0; j< DIGEST_LENGTH*2; j++)
						cout << d[j];
						cout<<endl;
					}
					delete[] pos;
					delete[] vos;
				}
				else if (queryType==1)
				{
					long numberofhash=0;
					long numberofnodes=0;
					tree->calculateVOsize(indexarray, 2, &numberofhash, &numberofnodes);
					cout<<"number of hashes: "<<numberofhash<<" number of nodes: "<<numberofnodes<<endl;
					totalHash+=numberofhash;
					totalNodeVerify+=numberofnodes;
				}

			}

			if (((count+1) % 1000) == 0 && count!=0)
				cerr << count+1 << endl;
			count++;
		}
		
		//tree->showTree();
		long treesize=tree->calculateTreeSize();
		totalHash=totalHash/count;
		totalNodeVerify=totalNodeVerify/count;
		cout<<"fanout= "<<capacity<<endl;
		cout<<"tree size= "<<treesize<<" bytes"<<endl;	
		cout<<"number of hashes "<<totalHash<<endl;
		cout<<"number of node verify "<<totalNodeVerify<<endl;
		cout<<"avg vo size "<<calVOsize(totalHash, totalNodeVerify,numelements)<<endl;	
		delete tree;
			
		ru.stop();
	
		cerr << "Operations: " << count << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;
	
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
